# -*- coding: utf-8 -*-

"""Chang Density Objects: Periodic Grid + Lattice / Atoms"""
import math
from abc import ABCMeta, abstractmethod
from typing import Dict, List, Union

import numpy as np
from monty.json import MSONable
from pymatgen import Lattice, Structure
from pymatgen.io.vasp import VolumetricData, Chgcar, Poscar
from pyrho.core.pgrid import PGrid
import numpy.typing as npt


class ChargeABC(metaclass=ABCMeta):
    @abstractmethod
    def get_reshaped_cell(
        self,
        sc_mat: npt.ArrayLike = ((1, 0, 0), (0, 1, 0), (0, 0, 1)),
        frac_shift: npt.ArrayLike = (0.0, 0.0, 0.0),
        grid_out: Union[List, int] = int(1e9),
    ) -> "ChargeABC":
        pass

    @abstractmethod
    def reorient_axis(self) -> None:
        pass


class ChargeDensity(PGrid, ChargeABC):
    def __init__(
        self, grid_data: np.ndarray, structure: Structure, normalization: str = "vasp",
    ):
        """
        Class that contains functions to featurize volumetic data with periodic
        boundary conditions.
        Make sure the data being stored is grid-independent

        Args:
            grid_data: Volumetric data to read in
            structure: Atomic structure corresponding to the charge density
            normalization: the normalization scheme:
                - 'vasp' sum of the data / number of grid points == number of electrons
        """
        self.structure = structure.copy()
        self.normalization = normalization
        if normalization[0].lower() == "n":
            """
            No rescaling
            """
            scaled_data = grid_data
        elif normalization[0].lower() == "v":
            """
            the standard charge density from VASP is given as (rho*V) such that:
            sum(rho)/NGRID = NELECT/UC_vol
            so the real rho is:
            rho = (rho*UC_vol)*NGRID/UC_vol/UC_vol
            where the second V account for the different number of electrons in
            different cells
            """
            scaled_data = grid_data / self.structure.volume
        else:
            raise NotImplementedError("Not a valid normalization scheme")
        super().__init__(grid_data=scaled_data, lattice=None)

    @property
    def rho(self) -> np.ndarray:
        """
        Alias for the grid data, which should be the true charge density
        """
        return self.grid_data

    @property
    def lattice(self) -> np.ndarray:  # type: ignore
        """
        Override the lattice definition in PGrid, this makes getting the reoriented charge density easier.
        """
        return self.structure.lattice.matrix

    @property
    def renormalized_data(self) -> np.ndarray:
        if self.normalization[0].lower() == "n":
            return self.grid_data
        if self.normalization[0].lower() == "v":
            return self.grid_data * self.structure.volume
        else:
            raise NotImplementedError("Charge density normalization scheme not implemented")

    @classmethod
    def from_pmg_volumetric_data(cls, vdata: VolumetricData, data_key="total") -> "ChargeDensity":
        """
        Read a single key from the data field of a VolumetricData object
        Args:
            vdata: The volumetric data object
            data_key: The key to read from in the data field

        Returns:
            ChargeDensity object
        """
        return cls(grid_data=vdata.data[data_key], structure=vdata.structure, normalization="vasp",)

    @classmethod
    def from_rho(cls, rho: np.ndarray, structure: Structure, normalization: str = "vasp"):
        new_obj = cls(grid_data=rho, structure=structure, normalization="none")
        new_obj.normalization = normalization
        return new_obj

    def reorient_axis(self) -> None:
        """
        Change the orgientation of the lattice vector so that:
        a points along the x-axis, b is in the xy-plane, c is in the positive-z halve of space
        """
        self.structure.lattice = Lattice.from_parameters(
            *self.structure.lattice.abc, *self.structure.lattice.angles, vesta=True
        )

    # def get_data_in_cube(self, s: float, ngrid: int) -> np.ndarray:
    #     """
    #     Return the charge density data sampled on a cube.
    #
    #     Args:
    #         s: side lengthy in angstroms
    #         ngrid: number of grid points in each direction
    #
    #     Returns:
    #         ndarray: regridded data in a ngrid x ngrid x ngrid array
    #
    #     """
    #     grid_out = [ngrid, ngrid, ngrid]
    #     target_sc_lat_vecs = np.eye(3, 3) * s
    #     sc_mat = np.linalg.inv(self.structure.lattice.matrix) @ target_sc_lat_vecs
    #     _, res = get_sc_interp(self.rho, sc_mat, grid_out)
    #     return res.reshape(grid_out)
    #

    def get_transformed_obj(
        self,
        sc_mat: npt.ArrayLike = ((1, 0, 0), (0, 1, 0), (0, 0, 1)),
        frac_shift: npt.ArrayLike = (0.0, 0.0, 0.0),
        grid_out: Union[List[int], int] = int(1e9),
        up_sample: int = 1,
    ) -> "ChargeDensity":
        """
        Modify the structure and data and return a new object containing the reshaped
        data
        Args:
            sc_mat: Matrix to create the new cell
            frac_shift: translation to be applied on the cell after the matrix
            transformation
            grid_out: density of the new grid, can also just take the desired
            dimension as a list.

        Returns:

        """
        new_structure = self.structure.copy()
        new_structure.translate_sites(list(range(len(new_structure))), -np.array(frac_shift))
        new_structure = new_structure * sc_mat

        # determine the output grid
        lengths = new_structure.lattice.abc
        if isinstance(grid_out, int):
            ngrid = grid_out / new_structure.volume
            mult = (np.prod(lengths) / ngrid) ** (1 / 3)
            grid_out = [int(math.floor(max(l_ / mult, 1))) for l_ in lengths]
        else:
            grid_out = grid_out

        new_rho = self.get_transformed_data(sc_mat, frac_shift, grid_out=grid_out, up_sample=up_sample)
        return ChargeDensity.from_rho(new_rho, new_structure, self.normalization)

    def get_reshaped_cell(
        self,
        sc_mat: npt.ArrayLike = ((1, 0, 0), (0, 1, 0), (0, 0, 1)),
        frac_shift: npt.ArrayLike = (0.0, 0.0, 0.0),
        grid_out: Union[List, int] = int(1e9),
        up_sample: int = 1,
    ) -> "ChargeDensity":
        return self.get_transformed_obj(sc_mat=sc_mat, frac_shift=frac_shift, grid_out=grid_out, up_sample=up_sample)

    #
    #     _, new_rho = get_sc_interp(self.rho, sc_mat, grid_sizes=grid_out)
    #     new_rho = new_rho.reshape(grid_out)
    #
    #     grid_shifts = [
    #         int(t * g) for t, g in zip(translation - np.round(translation), grid_out)
    #     ]
    #
    #     new_rho = roll_array(new_rho, grid_shifts)
    #     return self.__class__.from_rho(new_rho, new_structure)


class SpinChargeDensity(MSONable, ChargeABC):
    def __init__(self, chargeden_dict: Dict, aug_charge: Dict = None):
        """
        Wrapper class that parses multiple sets of grid data on the same lattice

        Args:
            chargeden_dict: A dictionary containing multiple charge density objects
                        typically in the format {'total' : ChargeDen1, 'diff' : ChargeDen2}
        """
        self.chargeden_dict = chargeden_dict
        self.aug_charge = aug_charge
        self._tmp_key = next(
            iter(self.chargeden_dict)
        )  # get one key in the dictionary to make writing the subsequent code easier

    @classmethod
    def from_pmg_volumetric_data(cls, vdata: VolumetricData, data_keys=("total", "diff")):
        chargeden_dict = {}
        data_aug = getattr(vdata, "data_aug", None)
        for k in data_keys:
            chargeden_dict[k] = ChargeDensity.from_pmg_volumetric_data(vdata, data_key=k)
        return cls(chargeden_dict, aug_charge=data_aug)

    @property
    def lattice(self) -> Lattice:
        return self.chargeden_dict[self._tmp_key].lattice

    def to_Chgcar(self) -> Chgcar:
        struct = self.chargeden_dict[self._tmp_key].structure
        data_ = {k: v.renormalized_data for k, v in self.chargeden_dict.items()}
        return Chgcar(Poscar(struct), data_, data_aug=self.aug_charge)

    def to_VolumetricData(self) -> VolumetricData:
        key_ = next(iter(self.chargeden_dict))
        struct = self.chargeden_dict[key_].structure
        data_ = {k: v.renormalized_data for k, v in self.chargeden_dict.items()}
        return VolumetricData(struct, data_)

    def get_reshaped_cell(
        self,
        sc_mat: npt.ArrayLike = ((1, 0, 0), (0, 1, 0), (0, 0, 1)),
        frac_shift: npt.ArrayLike = (0.0, 0.0, 0.0),
        grid_out: Union[List, int] = int(1e9),
        up_sample: int = 1,
    ) -> "SpinChargeDensity":
        new_spin_charge = {}
        for k, v in self.chargeden_dict.items():
            new_spin_charge[k] = v.get_reshaped_cell(sc_mat, frac_shift, grid_out)
        factor = int(
            new_spin_charge[self._tmp_key].structure.num_sites / self.chargeden_dict[self._tmp_key].structure.num_sites
        )
        new_aug = {}
        if self.aug_charge is not None:
            for k, v in self.aug_charge.items():
                new_aug[k] = multiply_aug(v, factor)
        return SpinChargeDensity(new_spin_charge, new_aug)

    def reorient_axis(self) -> None:
        for k, v in self.chargeden_dict:
            v.reorient_axis()


def multiply_aug(data_aug: List[str], factor: int) -> List[str]:
    """
    The original idea here was to use to to speed up some vasp calculations for
    supercells by initializing the entire CHGCAR file.
    The current code does not deal with transformation of the Augemetation charges after regridding.

    This is a naive way to multiply the Augmentation data in the CHGCAR,
    a real working implementation will require analysis of the PAW projection operators.
    However, even with such an implementation, the speed up will be minimal due to VASP's interal
    minimization algorithms.
    Args:
        data_aug: The original augmentation data from a CHGCAR
        factor: The multiplication factor (some integer number of times it gets repeated)
    Returns:
        List of strings for each line of the Augmentation data.
    """
    res = []  # type: List[str]
    cur_block = []  # type: List[str]
    cnt = 0
    for ll in data_aug:
        if "augmentation" in ll:
            if cur_block:
                for j in range(factor):
                    cnt += 1
                    cur_block[0] = f"augmentation occupancies{cnt:>4}{cur_block[0].split()[-1]:>4}\n"
                    res.extend(cur_block)
            cur_block = [ll]
        else:
            cur_block.append(ll)
    else:
        for j in range(factor):
            cnt += 1
            cur_block[0] = f"augmentation occupancies{cnt:>4}{cur_block[0].split()[-1]:>4}\n"
            res.extend(cur_block)
    return res



# -*- coding: utf-8 -*-

"""Python class for ND grid data volumetric data"""
from typing import List, Union, Tuple
import numpy as np
from monty.json import MSONable
from pyrho.core.utils import (
    get_sc_interp,
    interpolate_fourier,
    get_ucell_frac_fit_sphere,
)
from scipy.ndimage import convolve
import numpy.typing as npt


class PGrid(MSONable):
    def __init__(self, grid_data: np.ndarray, lattice: Union[np.ndarray, None] = None):
        """
        Base class for N-dimensional Regular period grid data.
        The core code should be valid in N-dimensions and not depend on pymatgen

        Args:
            grid_data: Data stored on the regular rid
            lattice: list of lattice vectors
        """
        if lattice is not None:  # type: ignore
            self.lattice = np.array(lattice)
        self.grid_data = grid_data
        self._dim = len(self.grid_data.shape)
        self.grid_shape = self.grid_data.shape
        self.ngridpts = np.prod(self.grid_shape)

    def get_transformed_data(
        self, sc_mat: npt.ArrayLike, frac_shift: npt.ArrayLike, grid_out: List[int], up_sample: int = 1,
    ) -> np.ndarray:
        """
        Apply a transformation to the grid data
        This function assumes that the data is fixed in place and the transformation
        is applied to the lattice vectors.
        Args:
            sc_mat: matrix transformation applied to the lattice vectors
            frac_shift: shift the lattice in fractional coordinates of the input cell
            up_sample: the factor to scale up the sampling of the grid data

        sc_mat  --->  [2, 1]    trans  --->  [0.1, 0.3]
                      [0, 1]
        new lattice vectors:
            a = [0.2, 0.4] --> [2.2, 1.4]
            b = [0.2, 0.4] --> [0.2, 1.1]

        Returns:
            transformed data

        """
        if up_sample == 1:
            interp_grid_data = self.grid_data
        else:
            interp_grid_data = interpolate_fourier(
                arr_in=self.grid_data, shape=[g_dim_ * up_sample for g_dim_ in self.grid_data.shape],
            )
        _, new_rho = get_sc_interp(interp_grid_data, sc_mat, grid_sizes=grid_out, origin=frac_shift)  # type: ignore
        new_rho = new_rho.reshape(grid_out)

        # TODO make this part of the original transformation
        # grid_shifts = [
        #     int(t * g) for t, g in zip(frac_shift - np.round(frac_shift), grid_out)
        # ]
        #
        # new_rho = roll_array(new_rho, grid_shifts)

        return new_rho

    def get_transformed_obj(
        self,
        sc_mat: Union[List[List[int]], np.ndarray],
        frac_shift: Union[np.ndarray, List[float], Tuple[float]],
        grid_out: List[int],
        up_sample: int = 1,
    ) -> "PGrid":
        """
        Get a new PGrid object for the new transformed data
        Args:
            sc_mat: matrix transformation applied to the lattice vectors
            frac_shift: shift the lattice in fractional coordinates of the output cell
            grid_out: The size of the output grid to interpolate on
            up_sample: the factor to scale up the sampling of the grid data

        Returns:
            New PGrid object
        """
        new_data = self.get_transformed_data(sc_mat, frac_shift, grid_out=grid_out, up_sample=up_sample)
        new_lattice = np.dot(sc_mat, self.lattice)
        return PGrid(grid_data=new_data, lattice=new_lattice)

    def gaussian_smear(self, sigma: float = 0.2, arr_in: np.ndarray = None) -> Tuple[np.ndarray, np.ndarray]:
        """
        Applies an isotropic Gaussian smear of width (standard deviation) r to
        the potential field. This is necessary to avoid finding paths through
        narrow minima or nodes that may exist in the field (although any
        potential or charge distribution generated from GGA should be
        relatively smooth anyway). The smearing obeys periodic
        boundary conditions at the edges of the cell.

        Args:
            sigma: Smearing width in cartesian coordinates, in the same units as the
            structure lattice vectors
            arr_in: input data array to smear, if None: smear self.grid_data
        """
        # get the dimension of the filter needed for 1 std dev of gaussian mask
        # Go 4 standard deviations away
        if arr_in is None:
            arr = self.grid_data
        else:
            arr = arr_in
        r_frac = get_ucell_frac_fit_sphere(lattice=self.lattice, r=sigma * 5)
        filter_shape = [
            int(np.ceil(itr_rf * itr_dim / 2) * 2)  # dimension of the filter should be even
            for itr_rf, itr_dim in zip(r_frac, arr.shape)
        ]

        filter_latt = np.array([(filter_shape[_] + 1) / (arr.shape[_] + 1) * self.lattice[_] for _ in range(self._dim)])

        # Get the fractional positions
        filter_frac_c = [np.linspace(0, 1, _, endpoint=False) for _ in filter_shape]
        frac_pos = np.meshgrid(*filter_frac_c, indexing="ij")
        frac_pos = [_.flatten() for _ in frac_pos]

        # convert to cartesian
        cart_pos = np.matmul(filter_latt.T, np.vstack(frac_pos))

        # Distance the center if 1d we make  this iterable
        tmp = sum(filter_latt)  # type: Union[float, np.ndarray]
        if isinstance(tmp, np.ndarray):
            mid_point = tmp / 2  # type: Union[List, np.ndarray]
        else:
            mid_point = [tmp / 2]
        disp2mid2 = [(i_coord.reshape(filter_shape) - mp_coord) ** 2 for mp_coord, i_coord in zip(mid_point, cart_pos)]
        dist2mid = np.sqrt(sum(disp2mid2))
        # make sure the mask is zero?
        mm = dist2mid <= sigma * 4
        gauss = np.exp(-1 / 2 * (dist2mid / sigma) ** 2) * mm
        gauss = gauss / gauss.sum()
        return convolve(input=arr, weights=gauss, mode="wrap"), gauss

    def lossy_smooth_compression(self, grid_out: List, smear_std: float = 0.2) -> np.ndarray:
        """
        Perform Fourier interpolation then Gaussian smoothing.
        the smoothing makes sure that simple operation like max and min filters still
        give the same results.

        Args:
            grid_out: desired output grid of the compressed data,

        Returns:
            Smoothed array
        """
        arr_interp = np.absolute(interpolate_fourier(self.grid_data, grid_out))
        if smear_std > 0:
            arr_interp, _ = self.gaussian_smear(arr_in=arr_interp, sigma=smear_std)
        return arr_interp



# -*- coding: utf-8 -*-

"""Fourier analysis functions."""

# Calculate the fourier transform
from typing import List, Callable
import numpy.typing as npt
from monty.json import MSONable
import numpy as np


class PFourier(MSONable):
    def __init__(self, fourier_data: np.ndarray, lattice: npt.ArrayLike):
        """
        Fourier transform defined for a periodic function. The representation must be
        SC agnostic, (i.e. if we double the size of the unit cell)

        Much of the analysis relies on the fact that numpy.ndarray.flatten puts the array in a particular order.
        The correct ordering is used but not strictly enforced by the present code so use with care.
        Also make sure the fourier data is generated in the correct way: with index='ij'

        Args:
            fourier_data: Complex data for each point in fourier_pos
            lattice: The corresponding real space lattice parameters
        """
        self.lattice = np.array(lattice)
        self.fourier_data = fourier_data
        self.shape = fourier_data.shape

    @property
    def fractional_reciprocal_pos(self) -> np.ndarray:
        """
        Return fft position  structured grid
        Assuming standard fft format A[N-1] = A[-1]
        """
        grid_vec = [np.linspace(0, 1, isize, endpoint=False) for isize in self.fourier_data.shape]
        frac_coords = np.meshgrid(*grid_vec, indexing="ij")
        return np.vstack([icoord.flatten() for icoord in frac_coords])

    @property
    def fft_pos_centered(self) -> np.ndarray:
        """
        Return the fft positions where the N-k is changed to -k
        """
        return np.array([ipos - np.round(ipos) for ipos in self.fractional_reciprocal_pos])

    @property
    def fft_pos_centered_s(self) -> np.ndarray:
        """
        Return the fft positions where the N-k is changed to -k
        """
        return np.array([ipos * self.fourier_data.shape[itr] for itr, ipos in enumerate(self.fft_pos_centered)])

    @property
    def fft_pos_centered_cartesian(self) -> List:
        """
        Return the fft positions where the N-k is changed to -k
        """
        return np.dot(self.reciprocal_lattice.T, self.fft_pos_centered)

    @property
    def fft_pos_centered_cartesian_s(self) -> List:
        """
        Return the fft positions where the N-k is changed to -k
        """
        return np.dot(self.reciprocal_lattice.T, self.fft_pos_centered_s)

    @property
    def reciprocal_lattice(self) -> np.ndarray:
        """
        Return the reciprocal lattice. Note that this is the standard
        reciprocal lattice used for solid state physics with a factor of 2 *
        pi.
        """
        return np.linalg.inv(self.lattice).T * 2 * np.pi

    @property
    def cartesian_reciprocal_pos(self) -> np.ndarray:
        """
        Get the list of reciprocal lattice points in cartesian coordinates
        """
        return np.dot(self.reciprocal_lattice.T, self.fractional_reciprocal_pos)

    # Helper functions open to input
    def get_points_dict(self, filter_pos: Callable = None, filter_val: Callable = None):
        """
        Return filtered data as a (position, fft_coefficient) pair
        Args:
            filter_val: Filter function applied to the fourier data values (Example: Only keep large Fourier weights)
            ftiler_pos: Filter function applied to the reciprocal positions (Example: low pass filter)
        """
        for cart_pos, val in zip(self.cartesian_reciprocal_pos.T, self.fourier_data.flatten()):
            if filter_val is not None and not filter_val(val):
                continue
            if filter_pos is not None and not filter_pos(cart_pos):
                continue
            yield cart_pos, val



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
from typing import List

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

requirements = []  # type: List[str]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3.7",
]

setup(
    author="Jimmy-Xuan Shen",
    author_email="jmmshn@gmail.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Tools for re-griding volumetric quantum chemistry data for machine-learning purposes.",
    install_requires=requirements,
    license="BSD license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    name="mp-pyrho",
    packages=find_packages("src"),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/materialsproject/pyRho",
    version="0.0.9",
    zip_safe=False,
)
